;;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;;
;;;; CMUCL 20b Patch 000: Fix issue with REALPART and IMAGPART for
;;;; sse2 builds.  This example illustrates the bug.
;;;;
;;;;    (defun zot-r (z)
;;;;      (declare (type (simple-array (complex double-float) (*)) z))
;;;;      (realpart (aref z 0)))
;;;;
;;;;    (let ((z (make-array 1 :element-type '(complex double-float)
;;;;			 :initial-element #c(42d0 -42d0))))
;;;;      (zot-r z))
;;;;
;;;;

(pushnew :cmucl-20b-patch-000 *features*)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (intl::install)
  (intl::translation-enable))

(in-package :x86)

#+sse2
(define-vop (realpart/complex-single-float)
  (:translate realpart)
  (:args (x :scs (complex-single-reg complex-single-stack descriptor-reg)))
  (:arg-types complex-single-float)
  (:results (r :scs (single-reg)))
  (:result-types single-float)
  (:temporary (:sc single-reg) temp)
  (:policy :fast-safe)
  (:note _N"complex float realpart")
  (:generator 3
    (sc-case x
      (complex-single-reg
       (cond ((location= r x)
	      (inst xorps temp temp)	; temp = 0|0|0|0
	      (inst movss temp x)	; temp = 0|0|0|x
	      (inst movss r temp))	; r = temp
	     (t
	      (inst xorps r r)		; temp = 0|0|0|0
	      (inst movss r x))))	; r = 0|0|0|x
      (complex-single-stack
       (inst movss r (ea-for-csf-real-stack x)))
      (descriptor-reg
       (inst movss r (ea-for-csf-real-desc x))))))

#+sse2
(define-vop (realpart/complex-double-float)
  (:translate realpart)
  (:args (x :scs (complex-double-reg complex-double-stack descriptor-reg)))
  (:arg-types complex-double-float)
  (:results (r :scs (double-reg)))
  (:result-types double-float)
  (:temporary (:sc double-reg) temp)
  (:policy :fast-safe)
  (:note "complex float realpart")
  (:generator 3
    (sc-case x
      (complex-double-reg
       (cond ((location= r x)
	      (inst xorpd temp temp)	; temp = 0|0
	      (inst movsd temp x)	; temp = 0|x
	      (inst movsd r temp))	; r = temp
	     (t
	      (inst xorpd r r)		; r = 0|0
	      (inst movsd r x))))	; r = 0|x
      (complex-double-stack
       (inst movsd r (ea-for-cdf-real-stack x)))
      (descriptor-reg
       (inst movsd r (ea-for-cdf-real-desc x))))))

#+sse2
(define-vop (imagpart/complex-double-float)
  (:translate imagpart)
  (:args (x :scs (complex-double-reg complex-double-stack descriptor-reg)))
  (:arg-types complex-double-float)
  (:results (r :scs (double-reg)))
  (:result-types double-float)
  (:temporary (:sc complex-double-reg) temp)
  (:policy :fast-safe)
  (:note _N"complex float imagpart")
  (:generator 3
    (sc-case x
      (complex-double-reg
       (cond ((location= r x)
	      (inst xorpd temp temp)	; temp = 0|0
	      (inst movhlps temp x)	; temp = 0|b
	      (inst movsd r temp))	; r = temp
	     (t
	      (inst xorpd r r)		; r = 0|0
	      (inst movhlps r x))))	; r = 0|b
      (complex-double-stack
       (inst movsd r (ea-for-cdf-imag-stack x)))
      (descriptor-reg
       (inst movsd r (ea-for-cdf-imag-desc x))))))

#+sse2
(macrolet
    ((convert-complex (trans op to from)
       (let ((name (symbolicate to "/" from))
	     (from-sc (symbolicate from "-REG"))
	     (from-type (symbolicate from "-FLOAT"))
	     (to-sc (symbolicate to "-REG"))
	     (to-type (symbolicate to "-FLOAT")))
	 `(define-vop (,name)
	   (:translate ,trans)
	   (:args (x :scs (,from-sc) :target r))
	   (:arg-types ,from-type)
	   (:results (r :scs (,to-sc)))
	   (:result-types ,to-type)
	   (:policy :fast-safe)
	   (:generator 1
	     ;; NOTE: We don't have 128-bit aligned objects, so we
	     ;; can't use the stack or descriptors here.
	     (inst ,op r x))))))
  (convert-complex %complex-double-float cvtps2pd complex-double complex-single)
  (convert-complex %complex-single-float cvtpd2ps complex-single complex-double))
